@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2016 Ping Identity Corporation
+ * Copyright (C) 2013-2017 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -45,28 +45,22 @@
  * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- * @Author: Hans Zandbelt - hzandbelt@pingidentity.com
+ * @Author: Hans Zandbelt - hans.zandbelt@zmartzone.eu
  */
 
 #ifndef MOD_AUTH_OPENIDC_H_
 #define MOD_AUTH_OPENIDC_H_
 
 #include <stdint.h>
-#include <openssl/evp.h>
-#include <apr_uri.h>
-#include <apr_uuid.h>
 #include <httpd.h>
 #include <http_core.h>
 #include <http_config.h>
 #include <mod_auth.h>
 
-#include "apr_memcache.h"
-#include "apr_shm.h"
-#include "apr_global_mutex.h"
-
-#include "jose/apr_jose.h"
-
+#include "jose.h"
 #include "cache/cache.h"
+#include "parse.h"
+#include <apr_uuid.h>
 
 #ifdef APLOG_USE_MODULE
 APLOG_USE_MODULE(auth_openidc);
@@ -133,17 +127,21 @@ APLOG_USE_MODULE(auth_openidc);
 #define OIDC_DISC_CB_PARAM "oidc_callback"
 /* parameter name of the OP provider selection in the discovery response */
 #define OIDC_DISC_OP_PARAM "iss"
+/* parameter name of the user URL in the discovery response */
+#define OIDC_DISC_USER_PARAM "disc_user"
 /* parameter name of the original URL in the discovery response */
 #define OIDC_DISC_RT_PARAM "target_link_uri"
+/* parameter name of the original method in the discovery response */
+#define OIDC_DISC_RM_PARAM "method"
 /* parameter name of login hint in the discovery response */
 #define OIDC_DISC_LH_PARAM "login_hint"
 /* parameter name of parameters that need to be passed in the authentication request */
 #define OIDC_DISC_AR_PARAM "auth_request_params"
 
 /* value that indicates to use server-side cache based session tracking */
-#define OIDC_SESSION_TYPE_22_SERVER_CACHE 0
+#define OIDC_SESSION_TYPE_SERVER_CACHE 0
 /* value that indicates to use client cookie based session tracking */
-#define OIDC_SESSION_TYPE_22_CLIENT_COOKIE 1
+#define OIDC_SESSION_TYPE_CLIENT_COOKIE 1
 
 /* nonce bytes length */
 #define OIDC_PROTO_NONCE_LENGTH 32
@@ -168,6 +166,13 @@ APLOG_USE_MODULE(auth_openidc);
 /* accept bearer token as a cookie parameter (PingAccess) */
 #define OIDC_OAUTH_ACCEPT_TOKEN_IN_COOKIE  8
 
+/* the hash key of the cookie name value in the list of options */
+#define OIDC_OAUTH_ACCEPT_TOKEN_IN_OPTION_COOKIE_NAME "cookie-name"
+
+/* introspection method options */
+#define OIDC_INTROSPECTION_METHOD_GET  "GET"
+#define OIDC_INTROSPECTION_METHOD_POST "POST"
+
 /* prefix of the cookie that binds the state in the authorization request/response to the browser */
 #define OIDCStateCookiePrefix  "mod_auth_openidc_state_"
 
@@ -190,6 +195,7 @@ APLOG_USE_MODULE(auth_openidc);
 
 /* define the parameter value for the "logout" request that indicates a GET-style logout call from the OP */
 #define OIDC_GET_STYLE_LOGOUT_PARAM_VALUE "get"
+#define OIDC_IMG_STYLE_LOGOUT_PARAM_VALUE "img"
 
 /* define the name of the cookie/parameter for CSRF protection */
 #define OIDC_CSRF_NAME "x_csrf"
@@ -201,10 +207,24 @@ APLOG_USE_MODULE(auth_openidc);
 #define OIDC_CACHE_SECTION_JWKS "jwks"
 #define OIDC_CACHE_SECTION_ACCESS_TOKEN "access_token"
 #define OIDC_CACHE_SECTION_PROVIDER "provider"
+#define OIDC_CACHE_SECTION_REQUEST_URI "request_uri"
 
-typedef enum {
-	AUTHENTICATE, PASS, RETURN401, RETURN410
-} unauthenticated_action;
+/* http methods */
+#define OIDC_METHOD_GET       "get"
+#define OIDC_METHOD_FORM_POST "form_post"
+
+/* the maximum size of data that we accept in a single POST value: 1MB */
+#define OIDC_MAX_POST_DATA_LEN 1024 * 1024
+
+#define OIDC_UNAUTH_AUTHENTICATE 1
+#define OIDC_UNAUTH_PASS         2
+#define OIDC_UNAUTH_RETURN401    3
+#define OIDC_UNAUTH_RETURN410    4
+
+#define OIDC_REQUEST_URI_CACHE_DURATION 30
+
+#define OIDC_USER_INFO_TOKEN_METHOD_HEADER 0
+#define OIDC_USER_INFO_TOKEN_METHOD_POST   1
 
 typedef struct oidc_jwks_uri_t {
 	const char *url;
@@ -226,6 +246,8 @@ typedef struct oidc_provider_t {
 	char *jwks_uri;
 	char *client_id;
 	char *client_secret;
+	char *token_endpoint_tls_client_key;
+	char *token_endpoint_tls_client_cert;
 
 	// the next ones function as global default settings too
 	int ssl_validate_server;
@@ -250,6 +272,8 @@ typedef struct oidc_provider_t {
 	char *userinfo_signed_response_alg;
 	char *userinfo_encrypted_response_alg;
 	char *userinfo_encrypted_response_enc;
+	int userinfo_token_method;
+	char *request_object;
 } oidc_provider_t ;
 
 typedef struct oidc_remote_user_claim_t {
@@ -261,6 +285,8 @@ typedef struct oidc_oauth_t {
 	int ssl_validate_server;
 	char *client_id;
 	char *client_secret;
+	char *introspection_endpoint_tls_client_key;
+	char *introspection_endpoint_tls_client_cert;
 	char *introspection_endpoint_url;
 	char *introspection_endpoint_method;
 	char *introspection_endpoint_params;
@@ -306,6 +332,8 @@ typedef struct oidc_cfg {
 	int session_type;
 	/* session cookie or persistent cookie */
 	int persistent_session_cookie;
+	/* session cookie chunk size */
+	int session_cookie_chunk_size;
 
 	/* pointer to cache functions */
 	oidc_cache_t *cache;
@@ -345,24 +373,9 @@ typedef struct oidc_cfg {
 
 	char *crypto_passphrase;
 
-	EVP_CIPHER_CTX *encrypt_ctx;
-	EVP_CIPHER_CTX *decrypt_ctx;
-} oidc_cfg;
+	int provider_metadata_refresh_interval;
 
-typedef struct oidc_dir_cfg {
-	/* (optional) external OP discovery page */
-	char *discover_url;
-	char *cookie_path;
-	char *cookie;
-	char *authn_header;
-	unauthenticated_action unauth_action;
-	apr_array_header_t *pass_cookies;
-	apr_byte_t pass_info_in_headers;
-	apr_byte_t pass_info_in_env_vars;
-	apr_byte_t oauth_accept_token_in;
-	apr_hash_t *oauth_accept_token_options;
-	int oauth_token_introspect_interval;
-} oidc_dir_cfg;
+} oidc_cfg;
 
 int oidc_check_user_id(request_rec *r);
 #if MODULE_MAGIC_NUMBER_MAJOR >= 20100714
@@ -373,40 +386,42 @@ int oidc_auth_checker(request_rec *r);
 void oidc_request_state_set(request_rec *r, const char *key, const char *value);
 const char*oidc_request_state_get(request_rec *r, const char *key);
 int oidc_handle_jwks(request_rec *r, oidc_cfg *c);
+apr_byte_t oidc_post_preserve_javascript(request_rec *r, const char *location, char **javascript, char **javascript_method);
 
 // oidc_oauth
 int oidc_oauth_check_userid(request_rec *r, oidc_cfg *c);
 
 // oidc_proto.c
-
+char *oidc_proto_peek_jwt_header(request_rec *r, const char *jwt, char **alg);
 int oidc_proto_authorization_request(request_rec *r, struct oidc_provider_t *provider, const char *login_hint, const char *redirect_uri, const char *state, json_t *proto_state, const char *id_token_hint, const char *code_challenge, const char *auth_request_params);
 apr_byte_t oidc_proto_is_post_authorization_response(request_rec *r, oidc_cfg *cfg);
 apr_byte_t oidc_proto_is_redirect_authorization_response(request_rec *r, oidc_cfg *cfg);
 apr_byte_t oidc_proto_refresh_request(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *rtoken, char **id_token, char **access_token, char **token_type, int *expires_in, char **refresh_token);
-apr_byte_t oidc_proto_resolve_userinfo(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *access_token, const char **response);
+apr_byte_t oidc_proto_resolve_userinfo(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *id_token_sub, const char *access_token, const char **response);
 apr_byte_t oidc_proto_account_based_discovery(request_rec *r, oidc_cfg *cfg, const char *acct, char **issuer);
-apr_byte_t oidc_proto_parse_idtoken(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *id_token, const char *nonce, apr_jwt_t **jwt, apr_byte_t is_code_flow);
+apr_byte_t oidc_proto_url_based_discovery(request_rec *r, oidc_cfg *cfg, const char *url, char **issuer);
+apr_byte_t oidc_proto_parse_idtoken(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *id_token, const char *nonce, oidc_jwt_t **jwt, apr_byte_t is_code_flow);
 int oidc_proto_javascript_implicit(request_rec *r, oidc_cfg *c);
 apr_array_header_t *oidc_proto_supported_flows(apr_pool_t *pool);
 apr_byte_t oidc_proto_flow_is_supported(apr_pool_t *pool, const char *flow);
 apr_byte_t oidc_proto_validate_authorization_response(request_rec *r, const char *response_type, const char *requested_response_mode, char **code, char **id_token, char **access_token, char **token_type, const char *used_response_mode);
-apr_byte_t oidc_proto_jwt_verify(request_rec *r, oidc_cfg *cfg, apr_jwt_t *jwt, const oidc_jwks_uri_t *jwks_uri, apr_hash_t *symmetric_keys);
-apr_byte_t oidc_proto_validate_jwt(request_rec *r, apr_jwt_t *jwt, const char *iss, apr_byte_t exp_is_mandatory, apr_byte_t iat_is_mandatory, int iat_slack);
+apr_byte_t oidc_proto_jwt_verify(request_rec *r, oidc_cfg *cfg, oidc_jwt_t *jwt, const oidc_jwks_uri_t *jwks_uri, apr_hash_t *symmetric_keys);
+apr_byte_t oidc_proto_validate_jwt(request_rec *r, oidc_jwt_t *jwt, const char *iss, apr_byte_t exp_is_mandatory, apr_byte_t iat_is_mandatory, int iat_slack);
 apr_byte_t oidc_proto_generate_nonce(request_rec *r, char **nonce, int len);
 apr_byte_t oidc_proto_generate_code_verifier(request_rec *r, char **code_verifier, int len);
 apr_byte_t oidc_proto_generate_code_challenge(request_rec *r, const char *code_verifier, char **code_challenge, const char *challenge_method);
 
-apr_byte_t oidc_proto_authorization_response_code_idtoken_token(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, apr_jwt_t **jwt);
-apr_byte_t oidc_proto_authorization_response_code_idtoken(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, apr_jwt_t **jwt);
-apr_byte_t oidc_proto_handle_authorization_response_code_token(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, apr_jwt_t **jwt);
-apr_byte_t oidc_proto_handle_authorization_response_code(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, apr_jwt_t **jwt);
-apr_byte_t oidc_proto_handle_authorization_response_idtoken_token(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, apr_jwt_t **jwt);
-apr_byte_t oidc_proto_handle_authorization_response_idtoken(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, apr_jwt_t **jwt);
+apr_byte_t oidc_proto_authorization_response_code_idtoken_token(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, oidc_jwt_t **jwt);
+apr_byte_t oidc_proto_authorization_response_code_idtoken(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, oidc_jwt_t **jwt);
+apr_byte_t oidc_proto_handle_authorization_response_code_token(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, oidc_jwt_t **jwt);
+apr_byte_t oidc_proto_handle_authorization_response_code(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, oidc_jwt_t **jwt);
+apr_byte_t oidc_proto_handle_authorization_response_idtoken_token(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, oidc_jwt_t **jwt);
+apr_byte_t oidc_proto_handle_authorization_response_idtoken(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, oidc_jwt_t **jwt);
 
 // non-static for test.c
-apr_byte_t oidc_proto_validate_access_token(request_rec *r, oidc_provider_t *provider, apr_jwt_t *jwt, const char *response_type, const char *access_token);
-apr_byte_t oidc_proto_validate_code(request_rec *r, oidc_provider_t *provider, apr_jwt_t *jwt, const char *response_type, const char *code);
-apr_byte_t oidc_proto_validate_nonce(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *nonce, apr_jwt_t *jwt);
+apr_byte_t oidc_proto_validate_access_token(request_rec *r, oidc_provider_t *provider, oidc_jwt_t *jwt, const char *response_type, const char *access_token);
+apr_byte_t oidc_proto_validate_code(request_rec *r, oidc_provider_t *provider, oidc_jwt_t *jwt, const char *response_type, const char *code);
+apr_byte_t oidc_proto_validate_nonce(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *nonce, oidc_jwt_t *jwt);
 
 // oidc_authz.c
 int oidc_authz_worker(request_rec *r, const json_t *const claims, const require_line *const reqs, int nelts);
@@ -421,32 +436,45 @@ void *oidc_merge_server_config(apr_pool_t *pool, void *BASE, void *ADD);
 void *oidc_create_dir_config(apr_pool_t *pool, char *path);
 void *oidc_merge_dir_config(apr_pool_t *pool, void *BASE, void *ADD);
 void oidc_register_hooks(apr_pool_t *pool);
+char *oidc_cfg_dir_discover_url(request_rec *r);
+char *oidc_cfg_dir_cookie(request_rec *r);
+char *oidc_cfg_dir_cookie_path(request_rec *r);
+char *oidc_cfg_dir_authn_header(request_rec *r);
+int oidc_cfg_dir_pass_info_in_headers(request_rec *r);
+int oidc_cfg_dir_pass_info_in_envvars(request_rec *r);
+int oidc_cfg_dir_pass_refresh_token(request_rec *r);
+int oidc_cfg_dir_accept_token_in(request_rec *r);
+char *oidc_cfg_dir_accept_token_in_option(request_rec *r, const char *key);
+int oidc_cfg_token_introspection_interval(request_rec *r);
+int oidc_cfg_dir_preserve_post(request_rec *r);
+apr_array_header_t *oidc_dir_cfg_pass_cookies(request_rec *r);
+apr_array_header_t *oidc_dir_cfg_strip_cookies(request_rec *r);
+int oidc_dir_cfg_unauth_action(request_rec *r);
+oidc_valid_function_t oidc_cfg_get_valid_endpoint_auth_function(oidc_cfg *cfg);
 
 // oidc_util.c
 int oidc_strnenvcmp(const char *a, const char *b, int len);
 int oidc_base64url_encode(request_rec *r, char **dst, const char *src, int src_len, int remove_padding);
-int oidc_base64url_decode(request_rec *r, char **dst, const char *src);
-int oidc_encrypt_base64url_encode_string(request_rec *r, char **dst, const char *src);
-int oidc_base64url_decode_decrypt_string(request_rec *r, char **dst, const char *src);
+int oidc_base64url_decode(apr_pool_t *pool, char **dst, const char *src);
 const char *oidc_get_current_url_host(request_rec *r);
 char *oidc_get_current_url(request_rec *r);
 char *oidc_url_encode(const request_rec *r, const char *str, const char *charsToEncode);
 char *oidc_normalize_header_name(const request_rec *r, const char *str);
-
 void oidc_util_set_cookie(request_rec *r, const char *cookieName, const char *cookieValue, apr_time_t expires);
 char *oidc_util_get_cookie(request_rec *r, const char *cookieName);
-apr_byte_t oidc_util_http_get(request_rec *r, const char *url, const apr_table_t *params, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy, apr_array_header_t *pass_cookies);
-apr_byte_t oidc_util_http_post_form(request_rec *r, const char *url, const apr_table_t *params, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy, apr_array_header_t *pass_cookies);
-apr_byte_t oidc_util_http_post_json(request_rec *r, const char *url, const json_t *data, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy, apr_array_header_t *pass_cookies);
+apr_byte_t oidc_util_http_get(request_rec *r, const char *url, const apr_table_t *params, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy, apr_array_header_t *pass_cookies, const char *ssl_cert, const char *ssl_key);
+apr_byte_t oidc_util_http_post_form(request_rec *r, const char *url, const apr_table_t *params, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy, apr_array_header_t *pass_cookies, const char *ssl_cert, const char *ssl_key);
+apr_byte_t oidc_util_http_post_json(request_rec *r, const char *url, const json_t *data, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy, apr_array_header_t *pass_cookies, const char *ssl_cert, const char *ssl_key);
 apr_byte_t oidc_util_request_matches_url(request_rec *r, const char *url);
 apr_byte_t oidc_util_request_has_parameter(request_rec *r, const char* param);
 apr_byte_t oidc_util_get_request_parameter(request_rec *r, char *name, char **value);
+apr_byte_t oidc_util_decode_json_object(request_rec *r, const char *str, json_t **json);
 apr_byte_t oidc_util_decode_json_and_check_error(request_rec *r, const char *str, json_t **json);
 int oidc_util_http_send(request_rec *r, const char *data, int data_len, const char *content_type, int success_rvalue);
 int oidc_util_html_send(request_rec *r, const char *title, const char *html_head, const char *on_load, const char *html_body, int status_code);
 char *oidc_util_escape_string(const request_rec *r, const char *str);
 char *oidc_util_unescape_string(const request_rec *r, const char *str);
-apr_byte_t oidc_util_read_form_encoded_params(request_rec *r, apr_table_t *table, const char *data);
+apr_byte_t oidc_util_read_form_encoded_params(request_rec *r, apr_table_t *table, char *data);
 apr_byte_t oidc_util_read_post_params(request_rec *r, apr_table_t *table);
 apr_byte_t oidc_util_file_read(request_rec *r, const char *path, apr_pool_t *pool, char **result);
 apr_byte_t oidc_util_issuer_match(const char *a, const char *b);
@@ -462,53 +490,38 @@ apr_byte_t oidc_json_object_get_string(apr_pool_t *pool, json_t *json, const cha
 apr_byte_t oidc_json_object_get_int(apr_pool_t *pool, json_t *json, const char *name, int *value, const int default_value);
 char *oidc_util_html_escape(apr_pool_t *pool, const char *input);
 void oidc_util_table_add_query_encoded_params(apr_pool_t *pool, apr_table_t *table, const char *params);
-apr_hash_t * oidc_util_merge_symmetric_key(apr_pool_t *pool, apr_hash_t *private_keys, const char *secret, const char *hash_algo);
 apr_hash_t * oidc_util_merge_key_sets(apr_pool_t *pool, apr_hash_t *k1, apr_hash_t *k2);
 apr_byte_t oidc_util_regexp_first_match(apr_pool_t *pool, const char *input, const char *regexp, char **output, char **error_str);
 apr_byte_t oidc_util_json_merge(json_t *src, json_t *dst);
 int oidc_util_cookie_domain_valid(const char *hostname, char *cookie_domain);
 apr_byte_t oidc_util_hash_string_and_base64url_encode(request_rec *r, const char *openssl_hash_algo, const char *input, char **output);
-
-// oidc_crypto.c
-unsigned char *oidc_crypto_aes_encrypt(request_rec *r, oidc_cfg *cfg, unsigned char *plaintext, int *len);
-unsigned char *oidc_crypto_aes_decrypt(request_rec *r, oidc_cfg *cfg, unsigned char *ciphertext, int *len);
-apr_byte_t oidc_crypto_destroy(oidc_cfg *cfg, server_rec *s);
+apr_byte_t oidc_util_jwt_create(request_rec *r, const char *secret, json_t *payload, char **compact_encoded_jwt);
+apr_byte_t oidc_util_jwt_verify(request_rec *r, const char *secret, const char *compact_encoded_jwt, json_t **result);
+char *oidc_util_get_chunked_cookie(request_rec *r, const char *cookieName, int cookie_chunk_size);
+void oidc_util_set_chunked_cookie(request_rec *r, const char *cookieName, const char *cookieValue, apr_time_t expires, int chunkSize);
+apr_byte_t oidc_util_create_symmetric_key(request_rec *r, const char *client_secret, int r_key_len, const char *hash_algo, apr_byte_t set_kid, oidc_jwk_t **jwk);
+apr_hash_t * oidc_util_merge_symmetric_key(apr_pool_t *pool, apr_hash_t *private_keys, oidc_jwk_t *jwk);
 
 // oidc_metadata.c
 apr_byte_t oidc_metadata_provider_retrieve(request_rec *r, oidc_cfg *cfg, const char *issuer, const char *url, json_t **j_metadata, const char **response);
-apr_byte_t oidc_metadata_provider_parse(request_rec *r, json_t *j_provider, oidc_provider_t *provider);
+apr_byte_t oidc_metadata_provider_parse(request_rec *r, oidc_cfg *cfg, json_t *j_provider, oidc_provider_t *provider);
 apr_byte_t oidc_metadata_list(request_rec *r, oidc_cfg *cfg, apr_array_header_t **arr);
 apr_byte_t oidc_metadata_get(request_rec *r, oidc_cfg *cfg, const char *selected, oidc_provider_t **provider, apr_byte_t allow_discovery);
 apr_byte_t oidc_metadata_jwks_get(request_rec *r, oidc_cfg *cfg, const oidc_jwks_uri_t *jwks_uri, json_t **j_jwks, apr_byte_t *refresh);
 
 // oidc_session.c
-#if MODULE_MAGIC_NUMBER_MAJOR_NOT_WORKING_YET >= 20081201
-// this stuff should make it easy to migrate to the post 2.3 mod_session infrastructure
-#include "mod_session.h"
-//#define OIDC_SESSION_USE_APACHE_SESSIONS 1
-#else
 typedef struct {
-    apr_pool_t *pool;             /* pool to be used for this session */
-    apr_uuid_t *uuid;             /* anonymous uuid of this particular session */
-    const char *remote_user;      /* user who owns this particular session */
-    apr_table_t *entries;         /* key value pairs */
-    const char *encoded;          /* the encoded version of the key value pairs */
-    apr_time_t expiry;            /* if > 0, the time of expiry of this session */
-    long maxage;                  /* if > 0, the maxage of the session, from
-                                   * which expiry is calculated */
-    int dirty;                    /* dirty flag */
-    int cached;                   /* true if this session was loaded from a
-                                   * cache of some kind */
-    int written;                  /* true if this session has already been
-                                   * written */
-} session_rec;
-#endif
+	char uuid[APR_UUID_FORMATTED_LENGTH + 1]; /* unique id */
+    const char *remote_user;                  /* user who owns this particular session */
+    json_t *state;                            /* the state for this session, encoded in a JSON object */
+    apr_time_t expiry;                        /* if > 0, the time of expiry of this session */
+} oidc_session_t;
 
-apr_status_t oidc_session_init();
-apr_status_t oidc_session_load(request_rec *r, session_rec **z);
-apr_status_t oidc_session_get(request_rec *r, session_rec *z, const char *key, const char **value);
-apr_status_t oidc_session_set(request_rec *r, session_rec *z, const char *key, const char *value);
-apr_status_t oidc_session_save(request_rec *r, session_rec *z);
-apr_status_t oidc_session_kill(request_rec *r, session_rec *z);
+apr_byte_t oidc_session_load(request_rec *r, oidc_session_t **z);
+apr_byte_t oidc_session_get(request_rec *r, oidc_session_t *z, const char *key, const char **value);
+apr_byte_t oidc_session_set(request_rec *r, oidc_session_t *z, const char *key, const char *value);
+apr_byte_t oidc_session_save(request_rec *r, oidc_session_t *z);
+apr_byte_t oidc_session_kill(request_rec *r, oidc_session_t *z);
+apr_byte_t oidc_session_free(request_rec *r, oidc_session_t *z);
 
 #endif /* MOD_AUTH_OPENIDC_H_ */
